#!/usr/bin/env python3

import cv2
# from cv2 import aruco
from tqdm import trange
import numpy as np
import os, os.path
from glob import glob
from collections import defaultdict
import pandas as pd

from .common import get_folders, true_basename, get_video_name
from .triangulate import load_offsets_dict, load_pose2d_fnames
from .compute_angles import get_angles

from aniposelib.cameras import CameraGroup

def get_transform(row):
    M = np.identity(3)
    center = np.zeros(3)
    for i in range(3):
        center[i] = np.mean(row['center_{}'.format(i)])
        for j in range(3):
            M[i, j] = np.mean(row['M_{}{}'.format(i, j)])
    return M, center

## TODO: handle missing cameras
def get_errors_group(config, group, scorer=None):
    if config['filter3d']['enabled']:
        pipeline_pose_3d = config['pipeline']['pose_3d_filter']
    else:
        pipeline_pose_3d = config['pipeline']['pose_3d']

    metadatas = dict()
    fnames_dict = dict()
    cam_names = []

    for cname, folder in group:
        metadata_fname = os.path.join('labeled-data', folder, 'anipose_metadata.csv')
        if scorer is None:
            labels_fname = sorted(glob(
                os.path.join('labeled-data', folder, 'CollectedData*.h5')))[0]
        else:
            labels_fname = os.path.join(
                'labeled-data', folder, 'CollectedData_{}.h5'.format(scorer))
            
        metadatas[cname] = pd.read_csv(metadata_fname)
        fnames_dict[cname] = labels_fname
        cam_names.append(cname)

    cam_names = sorted(cam_names)

    ## TODO: will have to modify this for custom offset per session
    offsets_dict = load_offsets_dict(config, cam_names)
    print(offsets_dict)

    out = load_pose2d_fnames(fnames_dict, offsets_dict, cam_names)

    points_labeled = out['points']
    bodyparts = out['bodyparts']

    metadata = metadatas[cam_names[0]]

    n_frames = len(metadata)
    n_joints = len(bodyparts)

    calib_fnames = np.array(metadata['calib'])
    points_3d_pred = np.full((n_frames, n_joints, 3), np.nan, 'float')
    points_3d_labeled = np.full((n_frames, n_joints, 3), np.nan, 'float')
    reproj_err_pred = np.full((n_frames, n_joints), np.nan, 'float')
    reproj_err_labeled = np.full((n_frames, n_joints), np.nan, 'float')

    # get predicted 3D points
    paths_3d = []
    curr_path = None
    curr_pose = None
    curr_fnum = None
    for i in range(n_frames):
        row = metadata.iloc[i]
        fname = row['video']
        fnum = row['framenum']
        prefix = os.path.dirname(os.path.dirname(fname))
        vidname = get_video_name(config, fname)
        pose_path = os.path.join(prefix, pipeline_pose_3d, vidname + '.csv')
        paths_3d.append(pose_path)
        if curr_path != pose_path:
            curr_pose = pd.read_csv(pose_path)
            curr_fnum = np.array(curr_pose['fnum'])
            curr_path = pose_path
        try:
            ix = np.where(curr_fnum == fnum)[0][0]
        except IndexError:
            print("W: frame {} not found in 3D data for video {}".format(fnum, fname))
            continue
        row = curr_pose.iloc[ix]
        M, center = get_transform(row)
        pts = np.array([(row[bp+'_x'], row[bp+'_y'], row[bp+'_z']) for bp in bodyparts])
        pts_t = (pts + center).dot(np.linalg.inv(M.T))
        points_3d_pred[i] = pts_t
        reproj_err_pred[i] = [row[bp + '_error'] for bp in bodyparts]
        
    # triangulate 3D points from labeled points
    # get reprojection errors as well
    curr_cgroup = None
    curr_calib_fname = None
    for i in range(n_frames):
        calib_fname = calib_fnames[i]
        if curr_calib_fname != calib_fname:
            print(calib_fname)
            curr_cgroup = CameraGroup.load(calib_fname)
            curr_cgroup = curr_cgroup.subset_cameras_names(cam_names)
            print(curr_cgroup.get_names())
            curr_calib_fname = calib_fname
        pts = points_labeled[:, i]
        p3d = curr_cgroup.triangulate(pts)
        points_3d_labeled[i] = p3d
        reproj_err_labeled[i] = curr_cgroup.reprojection_error(p3d, pts, mean=True)
        
    # get L2 and reprojection errors
    errors = np.linalg.norm(points_3d_labeled - points_3d_pred, axis=2)
    

    # get angles
    vecs_pred = dict()
    vecs_lab = dict()
    for bp_ix, bp in enumerate(bodyparts):
        vecs_lab[bp] = points_3d_labeled[:, bp_ix]
        vecs_pred[bp] = points_3d_pred[:, bp_ix]
    angles = config.get('angles', dict())
    # angle_names = sorted(angles.keys())
    angles_pred = get_angles(vecs_pred, angles)
    angles_lab = get_angles(vecs_lab, angles)
    angle_names = sorted(angles_pred.keys())

    # save into dataframe
    out = pd.DataFrame()
    out['pose_path'] = paths_3d
    out['framenum'] = metadata['framenum']
    out['calib'] = metadata['calib']
    out['img'] = metadata['img']
    out['video'] = metadata['video']
    for ang_name in angle_names:
        out[ang_name + '_lab'] = angles_lab[ang_name]
        out[ang_name + '_pred'] = angles_pred[ang_name]
        out[ang_name + '_error'] = angles_pred[ang_name] - angles_lab[ang_name]
    for bp_ix, bp in enumerate(bodyparts):
        out[bp + '_x_lab'] = points_3d_labeled[:, bp_ix, 0]
        out[bp + '_y_lab'] = points_3d_labeled[:, bp_ix, 1]
        out[bp + '_z_lab'] = points_3d_labeled[:, bp_ix, 2]
        out[bp + '_reprojerr_lab'] = reproj_err_labeled[:, bp_ix]
        out[bp + '_x_pred'] = points_3d_pred[:, bp_ix, 0]
        out[bp + '_y_pred'] = points_3d_pred[:, bp_ix, 1]
        out[bp + '_z_pred'] = points_3d_pred[:, bp_ix, 2]
        out[bp + '_reprojerr_pred'] = reproj_err_pred[:, bp_ix]
        out[bp + '_error'] = errors[:, bp_ix]
            
    return out

def get_tracking_errors(config, scorer=None):
    # pipeline_videos_raw = config['pipeline']['videos_raw']
    group_folders = defaultdict(list)
    folders = get_folders('labeled-data')

    for folder in folders:
        group, _, cname = folder.rpartition('--')
        group_folders[group].append( (cname, folder) )

    datas = []
    for group, ffs in group_folders.items():
        print(group)
        dd = get_errors_group(config, ffs)
        datas.append(dd)
    data = pd.concat(datas)

    outdir = os.path.join(config['path'],
                          config['pipeline']['summaries'])
    
    os.makedirs(outdir, exist_ok=True)
    
    data.to_csv(os.path.join(outdir, 'tracking_errors.csv'),
                index=False)

    print('Errors saved in {}'.format(
        os.path.join(outdir, 'tracking_errors.csv')))



#!/usr/bin/env python3

import numpy as np
from glob import glob
import pandas as pd
import os.path
import cv2
from tqdm import tqdm, trange
from collections import defaultdict
from scipy import signal
import queue
import threading

from aniposelib.cameras import CameraGroup

from .common import make_process_fun, get_nframes, \
    get_video_name, get_cam_name, \
    get_video_params, get_video_params_cap, \
    get_data_length, natural_keys, true_basename, find_calibration_folder

from .triangulate import load_offsets_dict
from .filter_pose import write_pose_2d

def get_projected_points(config, pose_fname, cgroup, offsets_dict):

    pose_data = pd.read_csv(pose_fname)
    cols = [x for x in pose_data.columns if '_error' in x]
    bodyparts = [c.replace('_error', '') for c in cols]

    M = np.identity(3)
    center = np.zeros(3)
    for i in range(3):
        center[i] = np.mean(pose_data['center_{}'.format(i)])
        for j in range(3):
            M[i, j] = np.mean(pose_data['M_{}{}'.format(i, j)])

    bp_dict = dict(zip(bodyparts, range(len(bodyparts))))

    all_points = np.array([np.array(pose_data.loc[:, (bp+'_x', bp+'_y', bp+'_z')])
                           for bp in bodyparts])

    all_errors = np.array([np.array(pose_data.loc[:, bp+'_error'])
                           for bp in bodyparts])

    all_scores = np.array([np.array(pose_data.loc[:, bp+'_score'])
                           for bp in bodyparts])
    
    if config['triangulation']['optim']:
        all_errors[np.isnan(all_errors)] = 0
    else:
        all_errors[np.isnan(all_errors)] = 10000
    good = (all_errors < 50)
    all_points[~good] = np.nan

    n_joints, n_frames, _ = all_points.shape
    n_cams = len(cgroup.cameras)

    all_points_flat = all_points.reshape(-1, 3)
    all_points_flat_t = (all_points_flat + center).dot(np.linalg.inv(M.T))

    points_2d_proj_flat = cgroup.project(all_points_flat_t)
    points_2d_proj = points_2d_proj_flat.reshape(n_cams, n_joints, n_frames, 2)

    cam_names = cgroup.get_names()
    for cix, cname in enumerate(cam_names):
        offset = offsets_dict[cname]
        dx, dy = offset[0], offset[1]
        points_2d_proj[cix, :, :, 0] -= dx
        points_2d_proj[cix, :, :, 1] -= dy

    return bodyparts, points_2d_proj, all_scores


def process_session(config, session_path):
    pipeline_videos_raw = config['pipeline']['videos_raw']
    pipeline_pose_3d = config['pipeline']['pose_3d']
    pipeline_pose_2d_projected = config['pipeline']['pose_2d_projected']

    video_ext = config['video_extension']

    vid_fnames_2d = glob(os.path.join(
        session_path, pipeline_videos_raw, "*."+video_ext))
    vid_fnames_2d = sorted(vid_fnames_2d, key=natural_keys)

    pose_fnames_3d = glob(os.path.join(
        session_path, pipeline_pose_3d, "*.csv"))
    pose_fnames_3d = sorted(pose_fnames_3d, key=natural_keys)
    
    if len(pose_fnames_3d) == 0:
        return

    fnames_2d = defaultdict(list)
    for vid in vid_fnames_2d:
        vidname = get_video_name(config, vid)
        fnames_2d[vidname].append(vid)

    fnames_3d = defaultdict(list)
    for fname in pose_fnames_3d:
        vidname = true_basename(fname)
        fnames_3d[vidname].append(fname)

    cgroup = None
    calib_folder = find_calibration_folder(config, session_path)
    if calib_folder is not None:
        calib_fname = os.path.join(calib_folder,
                                   config['pipeline']['calibration_results'],
                                   'calibration.toml')
        if os.path.exists(calib_fname):
            cgroup = CameraGroup.load(calib_fname)

    if cgroup is None:
        print('session {}: no calibration found, skipping'.format(session_path))
        return

    outdir = os.path.join(session_path, pipeline_pose_2d_projected)
    os.makedirs(outdir, exist_ok=True)

    for pose_fname in pose_fnames_3d:
        basename = true_basename(pose_fname)

        if len(fnames_2d[basename]) == 0:
            print(pose_fname, 'missing raw videos')
            continue

        fname_3d_current = pose_fname
        fnames_2d_current = fnames_2d[basename]
        fnames_2d_current = sorted(fnames_2d_current, key=natural_keys)

        out_fnames = [os.path.join(outdir, true_basename(fname) + '.h5')
                      for fname in fnames_2d_current]

        if all([os.path.exists(f) for f in out_fnames]):
            continue

        print(pose_fname)

        cam_names = [get_cam_name(config, fname)
                     for fname in fnames_2d_current]

        video_folder = os.path.join(session_path, pipeline_videos_raw)
        offsets_dict = load_offsets_dict(config, cam_names, video_folder)

        cgroup_subset = cgroup.subset_cameras_names(cam_names)

        bodyparts, points_2d_proj, all_scores = get_projected_points(
            config, fname_3d_current, cgroup_subset, offsets_dict)

        metadata = {
            'scorer': 'scorer',
            'bodyparts': bodyparts,
            'index': np.arange(points_2d_proj.shape[2])
        }

        n_cams, n_joints, n_frames, _ = points_2d_proj.shape
        
        pts = np.zeros((n_frames, n_joints, 3), dtype='float64')
        
        for cix, (cname, outname) in enumerate(zip(cam_names, out_fnames)):
            pts[:, :, :2] = points_2d_proj[cix].swapaxes(0, 1)
            pts[:, :, 2] = all_scores.T
            write_pose_2d(pts, metadata, outname)

project_2d_all = make_process_fun(process_session)



#!/usr/bin/env python3

import numpy as np
from glob import glob
import pandas as pd
import os.path
from tqdm import tqdm, trange
import sys
from collections import defaultdict
from scipy.spatial.transform import Rotation

from .common import make_process_fun, get_data_length, natural_keys


# project v onto u
def proj(u, v):
    return u * (np.sum(v * u, axis = 1) / np.sum(u * u, axis = 1))[:,None]


# othogonalize u with respect to v
def ortho(u, v):
    return u - proj(v, u)


def normalize(u):
    return u / np.linalg.norm(u, axis = 1)[:, None]


def get_angles(vecs, angles):
    outdict = dict()
    angle_names = list(angles.keys())
    for ang_name in angle_names:
        angle = angles[ang_name]
        if angle[0] == "chain":
            d = angles_chain(vecs, angle[1:])
            for k, v in d.items():
                outdict[k] = v
        elif len(angle) == 4 and angle[0] == 'axis':
            outdict[ang_name] = angles_axis(vecs, angle[1:])
        elif len(angle) == 4 and angle[0] == 'cross-axis':
            outdict[ang_name] = angles_crossaxis(vecs, angle[1:])
        else: # 'flex'
            outdict[ang_name] = angles_flex(vecs, angle[-3:])
    return outdict


def angles_chain(vecs, chain_list):
    chain = []
    flex_type = []
    for c in chain_list:
        if c[-1] == "/":
            chain.append(c[:-1])
            flex_type.append(-1)
        else:
            chain.append(c)
            flex_type.append(1)

    n_joints = len(chain)
    keypoints = np.array([vecs[c] for c in chain])
    n_points = keypoints.shape[1]

    xfs = []
    cc = Rotation.identity()
    xfs.append(cc)

    for i in range(n_joints-1):
        pos = keypoints[i+1]
        z_dir = normalize(pos - keypoints[i])
        if i == n_joints - 2: # pick an arbitrary axis for the last joint
            x_dir = ortho([1, 0, 0], z_dir)
            if np.linalg.norm(x_dir) < 1e-5:
                x_dir = ortho([0, 1, 0], z_dir)
        else:
            x_dir = ortho(keypoints[i+2] - pos, z_dir)
            x_dir *= flex_type[i+1]
        x_dir = normalize(x_dir)
        y_dir = np.cross(z_dir, x_dir)
        M = np.dstack([x_dir, y_dir, z_dir])
        rot = Rotation.from_matrix(M)
        xfs.append(rot)

    angles = []
    for i in range(n_joints-1):
        try:
            rot = xfs[i].inv() * xfs[i+1]
        except ValueError:
            print(" W: nan points, failed to compute angle chain")
            angles.append(np.full((n_points,3), np.nan))

        ang = rot.as_euler('zyx', degrees=True)
        if i != 0:
            flex = angles_flex(vecs, chain[i-1:i+2]) * flex_type[i]
            test = ~np.isclose(flex, ang[:,1])
            ang[:,0] += 180*test
            ang[:,1] = test*np.mod(-(ang[:,1]+180), 360) + (1-test)*ang[:,1]
            ang = np.mod(np.array(ang) + 180, 360) - 180
        angles.append(ang)



    outdict = dict()
    for i, (name, ang) in enumerate(zip(chain, angles)):
        outdict[name + "_flex"] = ang[:,1]
        if i != len(angles)-1:
            outdict[name + "_rot"] = ang[:,0]
        if i == 0:
            outdict[name + "_abduct"] = ang[:,2]

    return outdict


def angles_flex(vecs, angle):
    a,b,c = angle
    v1 = normalize(vecs[a] - vecs[b])
    v2 = normalize(vecs[c] - vecs[b])
    ang_rad = np.arccos(np.sum(v1 * v2, axis = 1))
    ang_deg = np.rad2deg(ang_rad)
    return ang_deg


def angles_axis(vecs, angle):    
    a,b,c = angle
    v1 = vecs[a] - vecs[b] 
    v2 = vecs[b] - vecs[c]
    z = normalize(v1)
    x = normalize(ortho([1, 0, 0], z))
    y = np.cross(z, x)
    ang_rad = np.arctan2(np.sum(v2 * y, axis = 1), np.sum(v2 * x, axis = 1))
    ang_deg = np.rad2deg(ang_rad)
    return ang_deg


def angles_crossaxis(vecs, angle):
    a,b,c = angle
    v1 = vecs[a] - vecs[b]
    v2 = vecs[b] -vecs[c]
    point = vecs[c] - vecs[a]
    z = normalize(np.cross(v1, v2))
    x = normalize(ortho([1, 0, 0], z))
    y = np.cross(z, x)
    ang_rad = np.arctan2(np.sum(point * y, axis = 1), np.sum(point * x, axis = 1))
    ang_deg = np.rad2deg(ang_rad)
    return ang_deg
    

def compute_angles(config, labels_fname, outname):
    data = pd.read_csv(labels_fname)

    cols = [x for x in data.columns if '_error' in x]
    bodyparts = [c.replace('_error', '') for c in cols]

    vecs = dict()
    for bp in bodyparts:
        vec = np.array(data[[bp+'_x', bp+'_y', bp+'_z']])
        vecs[bp] = vec
    
    outdict = get_angles(vecs, config.get('angles', dict()))
    outdict['fnum'] = data['fnum']
    
    dout = pd.DataFrame(outdict)
    dout.to_csv(outname, index=False)


def process_session(config, session_path):
    if 'angles' not in config: # don't process anything if no angles in config
        return
    
    if config['filter3d']['enabled']:
        pipeline_3d = config['pipeline']['pose_3d_filter']
    else:
        pipeline_3d = config['pipeline']['pose_3d']
    pipeline_angles = config['pipeline']['angles']

    labels_fnames = glob(os.path.join(session_path,
                                      pipeline_3d, '*.csv'))
    labels_fnames = sorted(labels_fnames, key=natural_keys)

    outdir = os.path.join(session_path, pipeline_angles)

    if len(labels_fnames) > 0:
        os.makedirs(outdir, exist_ok=True)

    for fname in labels_fnames:
        basename = os.path.basename(fname)
        basename = os.path.splitext(basename)[0]

        out_fname = os.path.join(outdir, basename+'.csv')

        if os.path.exists(out_fname):
            continue

        print(out_fname)

        compute_angles(config, fname, out_fname)


compute_angles_all = make_process_fun(process_session)



#!/usr/bin/env python3

import cv2
import re
import os
from collections import deque
from subprocess import check_output
import numpy as np

from aniposelib.boards import CharucoBoard, Checkerboard

def atoi(text):
    return int(text) if text.isdigit() else text

def nan_helper(y):
    return np.isnan(y), lambda z: z.nonzero()[0]

def natural_keys(text):
    '''
    alist.sort(key=natural_keys) sorts in human order
    http://nedbatchelder.com/blog/200712/human_sorting.html
    '''
    return [ atoi(c) for c in re.split('(\d+)', text) ]

def wc(filename):
    out = check_output(["wc", "-l", filename])
    num = out.decode('utf8').split(' ')[0]
    return int(num)

def get_data_length(fname):
    import pandas as pd
    try:
        numlines = wc(fname) - 1
    except:
        numlines = len(pd.read_csv(fname))
    return numlines

def get_video_params_cap(cap):
    params = dict()
    params['width'] = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
    params['height'] = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
    params['nframes'] = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    params['fps'] = cap.get(cv2.CAP_PROP_FPS)
    return params

def get_video_params(fname):
    cap = cv2.VideoCapture(fname)
    params = get_video_params_cap(cap)
    cap.release()
    return params

def get_folders(path):
    folders = next(os.walk(path))[1]
    return sorted(folders)


def true_basename(fname):
    basename = os.path.basename(fname)
    basename = os.path.splitext(basename)[0]
    return basename


def get_cam_name(config, fname):
    basename = true_basename(fname)

    cam_regex = config['triangulation']['cam_regex']
    match = re.search(cam_regex, basename)

    if not match:
        return None
    else:
        name = match.groups()[0]
        return name.strip()

def get_video_name(config, fname):
    basename = true_basename(fname)

    cam_regex = config['triangulation']['cam_regex']
    vidname = re.sub(cam_regex, '', basename)
    return vidname.strip()

# TODO: get rid of skvideo dependency
def get_duration(vidname):
    import skvideo.io
    metadata = skvideo.io.ffprobe(vidname)
    duration = float(metadata['video']['@duration'])
    return duration

def get_nframes(vidname):
    import skvideo.io
    try:
        metadata = skvideo.io.ffprobe(vidname)
        length = int(metadata['video']['@nb_frames'])
        return length
    except KeyError:
        return 0

def full_path(path):
    path_user = os.path.expanduser(path)
    path_full = os.path.abspath(path_user)
    path_norm = os.path.normpath(path_full)
    return path_norm

def split_full_path(path):
    out = []
    while path != '':
        new, cur = os.path.split(path)
        if cur != '':
            out.append(cur)
        if new == path:
            out.append(new)
            break
        path = new
    return list(reversed(out))


def process_all(config, process_session, **args):

    pipeline_prefix = config['path']
    nesting = config['nesting']

    output = dict()

    if nesting == 0:
        output[()] = process_session(config, pipeline_prefix, **args)
        return output

    folders = get_folders(pipeline_prefix)
    level = 1

    q = deque()

    next_folders = [ (os.path.join(pipeline_prefix, folder),
                      (folder,),
                      level)
                     for folder in folders ]
    q.extend(next_folders)

    while len(q) != 0:
        path, past_folders, level = q.pop()

        if nesting < 0:
            output[past_folders] = process_session(config, path, **args)

            folders = get_folders(path)
            next_folders = [ (os.path.join(path, folder),
                              past_folders + (folder,),
                              level+1)
                             for folder in folders ]
            q.extend(next_folders)
        else:
            if level == nesting:
                output[past_folders] = process_session(config, path, **args)
            elif level > nesting:
                continue
            elif level < nesting:
                folders = get_folders(path)
                next_folders = [ (os.path.join(path, folder),
                                  past_folders + (folder,),
                                  level+1)
                                 for folder in folders ]
                q.extend(next_folders)

    return output

def make_process_fun(process_session, **args):
    def fun(config):
        return process_all(config, process_session, **args)
    return fun

def find_calibration_folder(config, session_path):
    pipeline_calibration_videos = config['pipeline']['calibration_videos']
    nesting = config['nesting']

    # TODO: fix this for nesting = -1
    level = nesting
    curpath = session_path

    while level >= 0:
        checkpath = os.path.join(curpath, pipeline_calibration_videos)
        if os.path.isdir(checkpath):
            return curpath

        curpath = os.path.dirname(curpath)
        level -= 1



def get_calibration_board(config):

    calib = config['calibration']
    board_size = calib['board_size']
    board_type = calib['board_type'].lower()

    manual_verification = config['manual_verification']
    manually_verify = manual_verification['manually_verify']

    if board_type == 'aruco':
        raise NotImplementedError("aruco board is not implemented with the current pipeline")
    elif board_type == 'charuco':
        board = CharucoBoard(
            board_size[0], board_size[1],
            calib['board_square_side_length'],
            calib['board_marker_length'],
            calib['board_marker_bits'],
            calib['board_marker_dict_number'],
            manually_verify=manually_verify)



    elif board_type == 'checkerboard':
        board = Checkerboard(board_size[0], board_size[1],
                             calib['board_square_side_length'], manually_verify=manually_verify)
    else:
        raise ValueError("board_type should be one of "
                         "'aruco', 'charuco', or 'checkerboard' not '{}'".format(
                             board_type))

    return board


## TODO: support checkerboard drawing
def get_calibration_board_image(config):
    board = get_calibration_board(config)
    numx, numy = board.get_size()
    size = numx*200, numy*200
    img = board.draw(size)
    return img



#!/usr/bin/env python3

from tqdm import trange
import numpy as np
from collections import defaultdict
import os
import os.path
import pandas as pd
import toml
from numpy import array as arr
from glob import glob
from scipy import optimize
import cv2

from .common import make_process_fun, find_calibration_folder, \
    get_video_name, get_cam_name, natural_keys

from aniposelib.cameras import CameraGroup

def proj(u, v):
    """Project u onto v"""
    return u * np.dot(v,u) / np.dot(u,u)

def ortho(u, v):
    """Orthagonalize u with respect to v"""
    return u - proj(v, u)

def get_median(all_points_3d, ix):
    pts = all_points_3d[:, ix]
    pts = pts[~np.isnan(pts[:, 0])]
    return np.median(pts, axis=0)


def correct_coordinate_frame(config, all_points_3d, bodyparts):
    """Given a config and a set of points and bodypart names, this function will rotate the coordinate frame to match the one in config"""
    # import IPython
    # IPython.embed()
    bp_index = dict(zip(bodyparts, range(len(bodyparts))))
    axes_mapping = dict(zip('xyz', range(3)))

    ref_point = config['triangulation']['reference_point']
    axes_spec = config['triangulation']['axes']

    # a_dirx, a_l, a_r = axes_spec[0]
    # b_dirx, b_l, b_r = axes_spec[1]

    # a_dir = axes_mapping[a_dirx]
    # b_dir = axes_mapping[b_dirx]


    # a_lv = get_median(all_points_3d, bp_index[a_l])
    # a_rv = get_median(all_points_3d, bp_index[a_r])
    # b_lv = get_median(all_points_3d, bp_index[b_l])
    # b_rv = get_median(all_points_3d, bp_index[b_r])

    # a_diff = a_rv - a_lv
    # b_diff = ortho(b_rv - b_lv, a_diff)

    M = np.zeros((3,3))
    dirs = []
    for (a_dirx, a_l, a_r) in axes_spec:
        a_dir = axes_mapping[a_dirx]
        a_lv = get_median(all_points_3d, bp_index[a_l])
        a_rv = get_median(all_points_3d, bp_index[a_r])
        a_diff = a_rv - a_lv
        M[a_dir] += a_diff / np.linalg.norm(a_diff)
        if a_dir not in dirs:
            dirs.append(a_dir)

    a_dir = dirs[0]
    b_dir = dirs[1]

    ## find the missing direction
    done = np.zeros(3, dtype='bool')
    done[a_dir] = True
    done[b_dir] = True
    c_dir = np.where(~done)[0][0]

    a_diff = M[a_dir]
    b_diff = M[b_dir]

    # M = np.zeros((3,3))
    # M[a_dir] = a_diff
    # M[b_dir] = b_diff
    #
    # form a right handed coordinate system
    if (a_dir,b_dir) in [(0,1), (2,0), (1,2)]:
        M[c_dir] = np.cross(a_diff, b_diff)
    else:
        M[c_dir] = np.cross(b_diff, a_diff)

    M /= np.linalg.norm(M, axis=1)[:,None]

    center = get_median(all_points_3d, bp_index[ref_point])

    all_points_3d_adj = all_points_3d.dot(M.T)
    center_new = get_median(all_points_3d_adj, bp_index[ref_point])
    all_points_3d_adj = all_points_3d_adj - center_new

    return all_points_3d_adj, M, center_new


def load_pose2d_fnames(fname_dict, offsets_dict=None, cam_names=None):
    if cam_names is None:
        cam_names = sorted(fname_dict.keys())
    pose_names = [fname_dict[cname] for cname in cam_names]

    if offsets_dict is None:
        offsets_dict = dict([(cname, (0,0)) for cname in cam_names])

    datas = []
    for ix_cam, (cam_name, pose_name) in \
            enumerate(zip(cam_names, pose_names)):
        dlabs = pd.read_hdf(pose_name)
        if len(dlabs.columns.levels) > 2:
            scorer = dlabs.columns.levels[0][0]
            dlabs = dlabs.loc[:, scorer]

        bp_index = dlabs.columns.names.index('bodyparts')
        joint_names = list(dlabs.columns.get_level_values(bp_index).unique())
        dx = offsets_dict[cam_name][0]
        dy = offsets_dict[cam_name][1]

        for joint in joint_names:
            dlabs.loc[:, (joint, 'x')] += dx
            dlabs.loc[:, (joint, 'y')] += dy

        datas.append(dlabs)

    n_cams = len(cam_names)
    n_joints = len(joint_names)
    n_frames = min([d.shape[0] for d in datas])

    # frame, camera, bodypart, xy
    points = np.full((n_cams, n_frames, n_joints, 2), np.nan, 'float')
    scores = np.full((n_cams, n_frames, n_joints), np.zeros(1), 'float')#initialise as zeros, instead of NaN, makes more sense? 

    for cam_ix, dlabs in enumerate(datas):
        for joint_ix, joint_name in enumerate(joint_names):
            try:
                points[cam_ix, :, joint_ix] = np.array(dlabs.loc[:, (joint_name, ('x', 'y'))])[:n_frames] 
                scores[cam_ix, :, joint_ix] = np.array(dlabs.loc[:, (joint_name, ('likelihood'))])[:n_frames].ravel()
            except KeyError:
                pass

    return {
        'cam_names': cam_names,
        'points': points,
        'scores': scores,
        'bodyparts': joint_names
    }


def load_offsets_dict(config, cam_names, video_folder=None):
    ## TODO: make the recorder.toml file configurable
    # record_fname = os.path.join(video_folder, 'recorder.toml')

    # if os.path.exists(record_fname):
    #     record_dict = toml.load(record_fname)
    # else:
    #     record_dict = None
    #     # if 'cameras' not in config:
    #     # ## TODO: more detailed error?
    #     #     print("-- no crop windows found")
    #     #     return

    offsets_dict = dict()
    for cname in cam_names:
        # if record_dict is None:
        if 'cameras' not in config or cname not in config['cameras']:
            # print("W: no crop window found for camera {}, assuming no crop".format(cname))
            offsets_dict[cname] = (0, 0)
        else:
            offsets_dict[cname] = tuple(config['cameras'][cname]['offset'])
        # else:
        #     offsets_dict[cname] = record_dict['cameras'][cname]['video']['ROIPosition']

    return offsets_dict

def load_constraints(config, bodyparts, key='constraints'):
    constraints_names = config['triangulation'].get(key, [])
    bp_index = dict(zip(bodyparts, range(len(bodyparts))))
    constraints = []
    for a, b in constraints_names:
        assert a in bp_index, 'Bodypart {} from constraints not found in list of bodyparts'.format(a)
        assert b in bp_index, 'Bodypart {} from constraints not found in list of bodyparts'.format(b)
        con = [bp_index[a], bp_index[b]]
        constraints.append(con)
    return constraints


def triangulate(config,
                calib_folder, video_folder, pose_folder,
                fname_dict, output_fname):

    cam_names = sorted(fname_dict.keys())

    calib_fname = os.path.join(calib_folder, 'calibration.toml')
    cgroup = CameraGroup.load(calib_fname)

    offsets_dict = load_offsets_dict(config, cam_names, video_folder)

    out = load_pose2d_fnames(fname_dict, offsets_dict, cam_names)
    all_points_raw = out['points']
    all_scores = out['scores']
    bodyparts = out['bodyparts']

    cgroup = cgroup.subset_cameras_names(cam_names)

    n_cams, n_frames, n_joints, _ = all_points_raw.shape

    bad = all_scores < config['triangulation']['score_threshold']
    all_points_raw[bad] = np.nan

    if config['triangulation']['optim']:
        constraints = load_constraints(config, bodyparts)
        constraints_weak = load_constraints(config, bodyparts, 'constraints_weak')

        points_2d = all_points_raw
        scores_2d = all_scores

        points_shaped = points_2d.reshape(n_cams, n_frames*n_joints, 2)
        if config['triangulation']['ransac']:
            points_3d_init, _, _, _ = cgroup.triangulate_ransac(points_shaped, progress=True)
        else:
            points_3d_init = cgroup.triangulate(points_shaped, progress=True)
        points_3d_init = points_3d_init.reshape((n_frames, n_joints, 3))

        c = np.isfinite(points_3d_init[:, :, 0])
        if np.sum(c) < 20:
            print("warning: not enough 3D points to run optimization")
            points_3d = points_3d_init
        elif (not config['triangulation']['optim_chunking']) or \
             (config['triangulation']['optim_chunking_size'] + 100 > n_frames):
            points_3d = cgroup.optim_points(
                points_2d, points_3d_init,
                constraints=constraints,
                constraints_weak=constraints_weak,
                # scores=scores_2d,
                scale_smooth=config['triangulation']['scale_smooth'],
                scale_length=config['triangulation']['scale_length'],
                scale_length_weak=config['triangulation']['scale_length_weak'],
                n_deriv_smooth=config['triangulation']['n_deriv_smooth'],
                reproj_error_threshold=config['triangulation']['reproj_error_threshold'],
                verbose=True)
        else:
            # do optimization in chunks due to memory constraints
            chunk_size = config['triangulation']['optim_chunking_size']
            fixed = None
            points_3d_out = points_3d_init * np.nan
            for start in range(0, n_frames, chunk_size):
                if start == 0:
                    n_fixed = 0
                    p2d = points_2d[:, start:start+chunk_size]
                    p3d = points_3d_init[start:start+chunk_size]
                else:
                    n_fixed = 15
                    p2d = points_2d[:, start-n_fixed:start+chunk_size]
                    p3d = np.vstack([fixed, points_3d_init[start:start+chunk_size]])

                points_3d = cgroup.optim_points(
                    p2d, p3d,
                    constraints=constraints,
                    constraints_weak=constraints_weak,
                    # scores=scores_2d,
                    scale_smooth=config['triangulation']['scale_smooth'],
                    scale_length=config['triangulation']['scale_length'],
                    scale_length_weak=config['triangulation']['scale_length_weak'],
                    n_deriv_smooth=config['triangulation']['n_deriv_smooth'],
                    reproj_error_threshold=config['triangulation']['reproj_error_threshold'],
                    verbose=True, n_fixed=n_fixed)
                points_3d_out[start:start+chunk_size] = points_3d[n_fixed:]
                fixed = points_3d[-15:]
            points_3d = points_3d_out

        points_2d_flat = points_2d.reshape(n_cams, -1, 2)
        points_3d_flat = points_3d.reshape(-1, 3)

        errors = cgroup.reprojection_error(
            points_3d_flat, points_2d_flat, mean=True)
        good_points = ~np.isnan(all_points_raw[:, :, :, 0])
        num_cams = np.sum(good_points, axis=0).astype('float')

        all_points_3d = points_3d
        all_errors = errors.reshape(n_frames, n_joints)

        all_scores[~good_points] = 2
        scores_3d = np.min(all_scores, axis=0)

        scores_3d[num_cams < 1] = np.nan
        all_errors[num_cams < 1] = np.nan

    else:
        points_2d = all_points_raw.reshape(n_cams, n_frames*n_joints, 2)
        if config['triangulation']['ransac']:
            points_3d, picked, p2ds, errors = cgroup.triangulate_ransac(
                points_2d, min_cams=3, progress=True)

            all_points_picked = p2ds.reshape(n_cams, n_frames, n_joints, 2)
            good_points = ~np.isnan(all_points_picked[:, :, :, 0])

            num_cams = np.sum(np.sum(picked, axis=0), axis=1)\
                         .reshape(n_frames, n_joints)\
                         .astype('float')
        else:
            points_3d = cgroup.triangulate(points_2d, progress=True)
            errors = cgroup.reprojection_error(points_3d, points_2d, mean=True)
            good_points = ~np.isnan(all_points_raw[:, :, :, 0])
            num_cams = np.sum(good_points, axis=0).astype('float')

        all_points_3d = points_3d.reshape(n_frames, n_joints, 3)
        all_errors = errors.reshape(n_frames, n_joints)

        all_scores[~good_points] = 2
        scores_3d = np.min(all_scores, axis=0)

        scores_3d[num_cams < 2] = np.nan
        all_errors[num_cams < 2] = np.nan
        num_cams[num_cams < 2] = np.nan

    if 'reference_point' in config['triangulation'] and 'axes' in config['triangulation']:
        all_points_3d_adj, M, center = correct_coordinate_frame(config, all_points_3d, bodyparts)
    else:
        all_points_3d_adj = all_points_3d
        M = np.identity(3)
        center = np.zeros(3)

    dout = pd.DataFrame()
    for bp_num, bp in enumerate(bodyparts):
        for ax_num, axis in enumerate(['x','y','z']):
            dout[bp + '_' + axis] = all_points_3d_adj[:, bp_num, ax_num]
        dout[bp + '_error'] = all_errors[:, bp_num]
        dout[bp + '_ncams'] = num_cams[:, bp_num]
        dout[bp + '_score'] = scores_3d[:, bp_num]

    for i in range(3):
        for j in range(3):
            dout['M_{}{}'.format(i, j)] = M[i, j]

    for i in range(3):
        dout['center_{}'.format(i)] = center[i]

    dout['fnum'] = np.arange(n_frames)

    dout.to_csv(output_fname, index=False)


def process_session(config, session_path):
    pipeline_videos_raw = config['pipeline']['videos_raw']
    pipeline_calibration_results = config['pipeline']['calibration_results']
    pipeline_pose = config['pipeline']['pose_2d']
    pipeline_pose_filter = config['pipeline']['pose_2d_filter']
    pipeline_3d = config['pipeline']['pose_3d']

    calibration_path = find_calibration_folder(config, session_path)
    if calibration_path is None:
        return

    if config['filter']['enabled']:
        pose_folder = os.path.join(session_path, pipeline_pose_filter)
    else:
        pose_folder = os.path.join(session_path, pipeline_pose)

    calib_folder = os.path.join(calibration_path, pipeline_calibration_results)
    video_folder = os.path.join(session_path, pipeline_videos_raw)
    output_folder = os.path.join(session_path, pipeline_3d)

    pose_files = glob(os.path.join(pose_folder, '*.h5'))

    cam_videos = defaultdict(list)

    for pf in pose_files:
        name = get_video_name(config, pf)
        cam_videos[name].append(pf)

    vid_names = cam_videos.keys()
    vid_names = sorted(vid_names, key=natural_keys)

    if len(vid_names) > 0:
        os.makedirs(output_folder, exist_ok=True)

    for name in vid_names:
        fnames = cam_videos[name]
        cam_names = [get_cam_name(config, f) for f in fnames]
        fname_dict = dict(zip(cam_names, fnames))

        output_fname = os.path.join(output_folder, name + '.csv')

        print(output_fname)
        
        if os.path.exists(output_fname):
            continue


        try:
            triangulate(config,
                        calib_folder, video_folder, pose_folder,
                        fname_dict, output_fname)
        except ValueError:
            import traceback, sys
            traceback.print_exc(file=sys.stdout)
            

triangulate_all = make_process_fun(process_session)



#!/usr/bin/env python3

from sklearn.neural_network import MLPRegressor, MLPClassifier
import pandas as pd
import os.path
import numpy as np
from glob import glob
from ruamel.yaml import YAML
import pickle


def get_dataset_location(model_folder):
    config_fname = os.path.join(model_folder, 'config.yaml')

    yaml = YAML(typ='rt')
    with open(config_fname, 'r') as f:
        dlc_config = yaml.load(f)

    iternum = dlc_config['iteration']

    fname_pat = os.path.join(
        model_folder, 'training-datasets', 'iteration-'+str(iternum),
        '*', 'CollectedData_*.h5')
    fname = glob(fname_pat)[0]

    return fname

def load_pose_2d_training(fname):
    data_orig = pd.read_hdf(fname)
    scorer = data_orig.columns.levels[0][0]
    data = data_orig.loc[:, scorer]

    bp_index = data.columns.names.index('bodyparts')
    coord_index = data.columns.names.index('coords')
    bodyparts = list(data.columns.get_level_values(bp_index).unique())

    n_frames = len(data)
    n_joints = len(bodyparts)
    test = np.array(data).reshape(n_frames, n_joints, 2)

    bad = np.any(~np.isfinite(test), axis=2)
    test[bad] = np.nan

    metadata = {
        'bodyparts': bodyparts,
        'scorer': scorer,
        'index': data.index
    }

    return test, metadata

def generate_training_data(scores, n_iters=5):
    Xs = []
    ys = []

    for i in range(n_iters):
        scores_perturb = scores.copy()
        good = scores_perturb == 1
        scores_perturb[good] = np.random.normal(1, 0.3, size=np.sum(good))
        scores_perturb[~good] = np.random.normal(0, 0.3, size=np.sum(~good))
        flipped = np.random.uniform(size=good.shape) < 0.05
        scores_perturb = np.clip(scores_perturb, 0, 1)
        scores_perturb[flipped] = 1 - scores_perturb[flipped]
        Xs.append(scores_perturb)
        ys.append(scores)

    X = np.vstack(Xs)
    y = np.vstack(ys)

    return X, y

def train_mlp_classifier(X, y):
    hidden = int(X.shape[1] * 1.5)
    
    mlp = MLPClassifier(hidden_layer_sizes=(hidden),
                        verbose=2, max_iter=2000,
                        activation='tanh', tol=1e-5,
                        learning_rate='adaptive', solver='adam',
                        early_stopping=True)
    mlp.fit(X, y)

    return mlp


def save_mlp_classifier(mlp, fname):
    with open(fname, 'wb') as f:
        pickle.dump(mlp, f)
    print('autoencoder saved at:\n  {}'.format(fname))
 

def train_autoencoder(config):
    model_folder = config['model_folder']
    data_fname = get_dataset_location(model_folder)
    data, metadata = load_pose_2d_training(data_fname)
    n_frames, n_joints, _ = data.shape
    scores = np.ones((n_frames, n_joints), dtype='float64')
    bad = np.any(~np.isfinite(data), axis=2)
    scores[bad] = 0
    X, y = generate_training_data(scores)
    mlp = train_mlp_classifier(X, y)
    out_fname = os.path.join(config['path'], 'autoencoder.pickle')
    save_mlp_classifier(mlp, out_fname)

    
# model_folder = '/jellyfish/research/tuthill/hand-demo-dlc-TuthillLab-2019-08-05'

# config = {'model_folder': model_folder, 'path': model_folder}
# train_autoencoder(config)

# get dataset from deeplabcut folder
# generate augmented dataset to train autoencoder
# train MLP classifier
# save result



#!/usr/bin/env python3

from tqdm import tqdm, trange
import os.path, os
import numpy as np
import pandas as pd
from numpy import array as arr
from glob import glob
from scipy import signal, stats
from scipy.interpolate import splev, splrep
from scipy.spatial.distance import cdist
from scipy.spatial import cKDTree
from scipy.special import logsumexp
from collections import Counter
from multiprocessing import cpu_count
from multiprocessing import Pool, get_context
import pickle

from .common import make_process_fun, natural_keys


def nan_helper(y):
    return np.isnan(y), lambda z: z.nonzero()[0]


def remove_dups(pts, thres=7):
    tindex = np.repeat(np.arange(pts.shape[0])[:, None], pts.shape[1], axis=1)*100
    pts_ix = np.dstack([pts, tindex])
    pts_flat = pts_ix.reshape(-1, 3)
    check = np.all(np.isfinite(pts_flat), axis=1)
    ix_good = np.where(check)[0]
    tree = cKDTree(pts_flat[ix_good])

    shape = (pts.shape[0], pts.shape[1])
    pairs = tree.query_pairs(thres)

    if len(pairs) == 0:
        return pts

    indices = [ix_good[b] for a, b in pairs]
    i0, i1 = np.unravel_index(indices, shape)
    pts_out = np.copy(pts)
    pts_out[i0, i1] = np.nan

    return pts_out

def viterbi_path(points, scores, n_back=3, thres_dist=30):
    n_frames = points.shape[0]

    points_nans = remove_dups(points, thres=5)
    # points_nans[scores < 0.01] = np.nan

    num_points = np.sum(~np.isnan(points_nans[:, :, 0]), axis=1)
    num_max = np.max(num_points)

    particles = np.zeros((n_frames, num_max * n_back + 1, 3), dtype='float64')
    valid = np.zeros(n_frames, dtype='int64')
    for i in range(n_frames):
        s = 0
        for j in range(n_back):
            if i-j < 0:
                break
            ixs = np.where(~np.isnan(points_nans[i-j, :, 0]))[0]
            n_valid = len(ixs)
            particles[i, s:s+n_valid, :2] = points[i-j, ixs]
            particles[i, s:s+n_valid, 2] = scores[i-j, ixs] * np.power(2.0, -j)
            s += n_valid
        if s == 0:
            particles[i, 0] = [-1, -1, 0.001] # missing point
            s = 1
        valid[i] = s

    ## viterbi algorithm
    n_particles = np.max(valid)

    T_logprob = np.zeros((n_frames, n_particles), dtype='float64')
    T_logprob[:] = -np.inf
    T_back = np.zeros((n_frames, n_particles), dtype='int64')

    T_logprob[0, :valid[0]] = np.log(particles[0, :valid[0], 2])
    T_back[0, :] = -1

    for i in range(1, n_frames):
        va, vb = valid[i-1], valid[i]
        pa = particles[i-1, :va, :2]
        pb = particles[i, :vb, :2]

        dists = cdist(pa, pb)
        cdf_high = stats.norm.logcdf(dists + 2, scale=thres_dist)
        cdf_low = stats.norm.logcdf(dists - 2, scale=thres_dist)
        cdfs = np.array([cdf_high, cdf_low])
        P_trans = logsumexp(cdfs.T, b=[1,-1], axis=2)

        P_trans[P_trans < -100] = -100

        # take care of missing transitions
        P_trans[pb[:, 0] == -1, :] = np.log(0.001)
        P_trans[:, pa[:, 0] == -1] = np.log(0.001)

        pflat = particles[i, :vb, 2]
        possible = T_logprob[i-1, :va] + P_trans

        T_logprob[i, :vb] = np.max(possible, axis=1) + np.log(pflat)
        T_back[i, :vb] = np.argmax(possible, axis=1)

    out = np.zeros(n_frames, dtype='int')
    out[-1] = np.argmax(T_logprob[-1])

    for i in range(n_frames-1, 0, -1):
        out[i-1] = T_back[i, out[i]]

    trace = [particles[i, out[i]] for i in range(n_frames)]
    trace = np.array(trace)

    points_new = trace[:, :2]
    scores_new = trace[:, 2]
    # scores_new[out >= num_points] = 0

    return points_new, scores_new


def viterbi_path_wrapper(args):
    jix, pts, scs, max_offset, thres_dist = args
    pts_new, scs_new = viterbi_path(pts, scs, max_offset, thres_dist)
    return jix, pts_new, scs_new


def load_pose_2d(fname):
    data_orig = pd.read_hdf(fname)
    scorer = data_orig.columns.levels[0][0]
    data = data_orig.loc[:, scorer]

    bp_index = data.columns.names.index('bodyparts')
    coord_index = data.columns.names.index('coords')
    bodyparts = list(data.columns.get_level_values(bp_index).unique())
    n_possible = len(data.columns.levels[coord_index])//3

    n_frames = len(data)
    n_joints = len(bodyparts)
    test = np.array(data).reshape(n_frames, n_joints, n_possible, 3)

    metadata = {
        'bodyparts': bodyparts,
        'scorer': scorer,
        'index': data.index
    }

    return test, metadata

def filter_pose_viterbi(config, all_points, bodyparts):
    n_frames, n_joints, n_possible, _ = all_points.shape

    points_full = all_points[:, :, :, :2]
    scores_full = all_points[:, :, :, 2]

    points_full[scores_full < config['filter']['score_threshold']] = np.nan

    points = np.full((n_frames, n_joints, 2), np.nan, dtype='float64')
    scores = np.empty((n_frames, n_joints), dtype='float64')

    if config['filter']['multiprocessing']:
        n_proc_default = max(min(cpu_count() // 2, n_joints), 1)
        n_proc = config['filter'].get('n_proc', n_proc_default)
    else:
        n_proc = 1
    ctx = get_context('spawn')
    pool = ctx.Pool(n_proc)

    max_offset = config['filter']['n_back']
    thres_dist = config['filter']['offset_threshold']

    iterable = [ (jix, points_full[:, jix, :], scores_full[:, jix],
                  max_offset, thres_dist)
                 for jix in range(n_joints) ]

    results = pool.imap_unordered(viterbi_path_wrapper, iterable)

    for jix, pts_new, scs_new in tqdm(results, ncols=70):
        points[:, jix] = pts_new
        scores[:, jix] = scs_new

    pool.close()
    pool.join()

    return points, scores


def write_pose_2d(all_points, metadata, outname=None):
    points = all_points[:, :, :2]
    scores = all_points[:, :, 2]

    scorer = metadata['scorer']
    bodyparts = metadata['bodyparts']
    index = metadata['index']

    columns = pd.MultiIndex.from_product(
        [[scorer], bodyparts, ['x', 'y', 'likelihood']],
        names=['scorer', 'bodyparts', 'coords'])

    dout = pd.DataFrame(columns=columns, index=index)

    dout.loc[:, (scorer, bodyparts, 'x')] = points[:, :, 0]
    dout.loc[:, (scorer, bodyparts, 'y')] = points[:, :, 1]
    dout.loc[:, (scorer, bodyparts, 'likelihood')] = scores

    dout = dout.infer_objects()  # need this to have floats not objects in newer pandas

    if outname is not None:
        dout.to_hdf(outname, 'df_with_missing', format='table', mode='w')

    return dout


def filter_pose_medfilt(config, all_points, bodyparts):
    n_frames, n_joints, n_possible, _ = all_points.shape

    points_full = all_points[:, :, :, :2]
    scores_full = all_points[:, :, :, 2]

    points = np.full((n_frames, n_joints, 2), np.nan, dtype='float64')
    scores = np.empty((n_frames, n_joints), dtype='float64')

    for bp_ix, bp in enumerate(bodyparts):
        x = points_full[:, bp_ix, 0, 0]
        y = points_full[:, bp_ix, 0, 1]
        score = scores_full[:, bp_ix, 0]

        xmed = signal.medfilt(x, kernel_size=config['filter']['medfilt'])
        ymed = signal.medfilt(y, kernel_size=config['filter']['medfilt'])

        errx = np.abs(x - xmed)
        erry = np.abs(y - ymed)
        err = errx + erry

        bad = np.zeros(len(x), dtype='bool')
        bad[err >= config['filter']['offset_threshold']] = True
        bad[score < config['filter']['score_threshold']] = True

        Xf = arr([x,y]).T
        Xf[bad] = np.nan

        Xfi = np.copy(Xf)

        for i in range(Xf.shape[1]):
            vals = Xfi[:, i]
            nans, ix = nan_helper(vals)
            # some data missing, but not too much
            if np.sum(nans) > 0 and np.mean(~nans) > 0.5 and np.sum(~nans) > 5:
                if config['filter']['spline']:
                    spline = splrep(ix(~nans), vals[~nans], k=3, s=0)
                    vals[nans]= splev(ix(nans), spline)
                else:
                    vals[nans] = np.interp(ix(nans), ix(~nans), vals[~nans])
            Xfi[:,i] = vals

        points[:, bp_ix, 0] = Xfi[:, 0]
        points[:, bp_ix, 1] = Xfi[:, 1]
        # dout[scorer, bp, 'interpolated'] = np.isnan(Xf[:, 0])

    scores = scores_full[:, :, 0]

    return points, scores

def filter_pose_autoencoder_scores(config, all_points, bodyparts):
    n_frames, n_joints, n_possible, _ = all_points.shape

    points_full = all_points[:, :, :, :2]
    scores_full = all_points[:, :, :, 2]

    scores_test = all_points[:, :, 0, 2]

    fname_model = config['filter']['autoencoder_path']
    with open(fname_model, 'rb') as f:
        mlp = pickle.load(f)

    scores_pred = mlp.predict_proba(scores_test)
    scores_pred_rep = np.repeat(scores_pred, n_possible, axis=1).reshape(scores_full.shape)

    scores_fixed = np.min([scores_pred_rep, scores_full], axis=0)

    return points_full, scores_fixed


def wrap_input(points, mean, std):
    pts_demean = (points - mean) / std
    pts_demean[~np.isfinite(pts_demean)] = 0
    # pts_demean = pts_demean - np.median(pts_demean, axis=1)[:, None]
    n_frames = pts_demean.shape[0]
    return pts_demean.reshape(n_frames, -1)
    # return np.hstack([pts_demean.reshape(n_frames, -1), scores])

def unwrap_input(X, mean, std):
    n_joints = X.shape[1] // 2
    pts_demean = X[:, :n_joints*2].reshape(-1, n_joints, 2)
    points = pts_demean * std + mean
    return points


def filter_pose_autoencoder_points(config, all_points, bodyparts):
    n_frames, n_joints, n_possible, _ = all_points.shape

    points_full = all_points[:, :, :, :2]
    scores_full = all_points[:, :, :, 2]

    points_test = all_points[:, :, 0, :2]
    scores_test = all_points[:, :, 0, 2]
    points_test[scores_test < 0.4] = np.nan

    fname_model = config['filter']['autoencoder_points_path']
    with open(fname_model, 'rb') as f:
        d = pickle.load(f)
    mlp = d['mlp']
    thres_low = d['thres_low']
    thres_lh = d['thres_lh']
    mean = d['mean']
    std = d['std']
    
    points_pred = unwrap_input(
        mlp.predict(wrap_input(
            points_test, mean, std)),
        mean, std)
    dist = np.linalg.norm(points_pred - points_test, axis=2)
    scores_pred = np.exp(-(dist - thres_low)/(thres_lh/2.3))
    scores_pred = np.clip(scores_pred, 0, 1)
    c = ~np.isfinite(scores_pred)
    scores_pred[c] = scores_test[c]

    scores_pred_rep = np.repeat(scores_pred, n_possible, axis=1).reshape(scores_full.shape)
    scores_fixed = np.min([scores_pred_rep, scores_full], axis=0)

    return points_full, scores_fixed

def wrap_points(points, scores):
    if len(points.shape) == 3: # n_possible = 1
        points = points[:, :, None]
        scores = scores[:, :, None]

    n_frames, n_joints, n_possible, _ = points.shape

    all_points = np.full((n_frames, n_joints, n_possible, 3), np.nan, dtype='float64')
    all_points[:, :, :, :2] = points
    all_points[:, :, :, 2] = scores

    return all_points


FILTER_MAPPING = {
    'medfilt': filter_pose_medfilt,
    'viterbi': filter_pose_viterbi,
    'autoencoder': filter_pose_autoencoder_scores,
    'autoencoder_points': filter_pose_autoencoder_points
}

POSSIBLE_FILTERS = FILTER_MAPPING.keys()

def process_session(config, session_path):
    pipeline_pose = config['pipeline']['pose_2d']
    pipeline_pose_filter = config['pipeline']['pose_2d_filter']
    filter_types = config['filter']['type']
    if not isinstance(filter_types, list):
        filter_types = [filter_types]

    for filter_type in filter_types:
        assert filter_type in POSSIBLE_FILTERS, \
            "Invalid filter type, should be one of {}, but found {}".format(POSSIBLE_FILTERS, filter_type)

    pose_folder = os.path.join(session_path, pipeline_pose)
    output_folder = os.path.join(session_path, pipeline_pose_filter)

    pose_files = glob(os.path.join(pose_folder, '*.h5'))
    pose_files = sorted(pose_files, key=natural_keys)

    if len(pose_files) > 0:
        os.makedirs(output_folder, exist_ok=True)

    for fname in pose_files:
        basename = os.path.basename(fname)
        outpath = os.path.join(session_path,
                               pipeline_pose_filter,
                               basename)

        if os.path.exists(outpath):
            continue

        print(outpath)
        all_points, metadata = load_pose_2d(fname)

        for filter_type in filter_types:
            filter_fun = FILTER_MAPPING[filter_type]
            points, scores = filter_fun(config, all_points, metadata['bodyparts'])
            all_points = wrap_points(points, scores)

        write_pose_2d(all_points[:, :, 0], metadata, outpath)


filter_pose_all = make_process_fun(process_session)



#!/usr/bin/env python3

from tqdm import tqdm, trange
import os.path, os
import numpy as np
import pandas as pd
from numpy import array as arr
from glob import glob
from scipy import signal
from scipy.interpolate import splev, splrep

from .common import make_process_fun, natural_keys

def medfilt_data(values, size=15):
    padsize = size+5
    vpad = np.pad(values, (padsize, padsize),
                  mode='median',
                  stat_length=5)
    vpadf = signal.medfilt(vpad, kernel_size=size)
    return vpadf[padsize:-padsize]

def nan_helper(y):
    return np.isnan(y), lambda z: z.nonzero()[0]

def interpolate_data(vals):
    nans, ix = nan_helper(vals)
    out = np.copy(vals)
    if np.mean(nans) > 0.85:
        return out
    out[nans] = np.interp(ix(nans), ix(~nans), vals[~nans])
    return out

def filter_pose(config, fname, outname):
    data = pd.read_csv(fname)

    cols = [x for x in data.columns if '_error' in x]
    bodyparts = [c.replace('_error', '') for c in cols]

    for bp in bodyparts:
        error = np.array(data[bp + '_error'])
        error[np.isnan(error)] = 100000
        bad = error > config['filter3d']['offset_threshold']
        for v in 'xyz':
            key = '{}_{}'.format(bp, v)
            values = np.array(data[key])
            values[bad] = np.nan
            values_intp = interpolate_data(values)
            values_filt = medfilt_data(values_intp,
                                       size=config['filter3d']['medfilt'])
            data[key] = values_filt
        data[bp+'_error'] = 10 # FIXME: hack for plotting
        
    data.to_csv(outname, index=False)


def process_session(config, session_path):
    pipeline_pose = config['pipeline']['pose_3d']
    pipeline_pose_filter = config['pipeline']['pose_3d_filter']

    pose_folder = os.path.join(session_path, pipeline_pose)
    output_folder = os.path.join(session_path, pipeline_pose_filter)

    pose_files = glob(os.path.join(pose_folder, '*.csv'))
    pose_files = sorted(pose_files, key=natural_keys)

    if len(pose_files) > 0:
        os.makedirs(output_folder, exist_ok=True)

    for fname in pose_files:
        basename = os.path.basename(fname)
        outpath = os.path.join(session_path,
                               pipeline_pose_filter,
                               basename)

        if os.path.exists(outpath):
            continue

        print(outpath)
        filter_pose(config, fname, outpath)


filter_pose_3d_all = make_process_fun(process_session)



#!/usr/bin/env python3

from tqdm import tqdm
import numpy as np
import os
from glob import glob
from collections import defaultdict
import pickle

from .common import \
    find_calibration_folder, make_process_fun, process_all, \
    get_cam_name, get_video_name, \
    get_calibration_board, split_full_path, get_video_params

from .triangulate import load_pose2d_fnames, load_offsets_dict

from aniposelib.cameras import CameraGroup

def get_pose2d_fnames(config, session_path):
    if config['filter']['enabled']:
        pipeline_pose = config['pipeline']['pose_2d_filter']
    else:
        pipeline_pose = config['pipeline']['pose_2d']
    fnames = glob(os.path.join(session_path, pipeline_pose, '*.h5'))
    return session_path, fnames


def load_2d_data(config, calibration_path):
    # start_path, _ = os.path.split(calibration_path)
    start_path = calibration_path

    nesting_path = len(split_full_path(config['path']))
    nesting_start = len(split_full_path(start_path))
    new_nesting = config['nesting'] - (nesting_start - nesting_path)

    new_config = dict(config)
    new_config['path'] = start_path
    new_config['nesting'] = new_nesting

    pose_fnames = process_all(new_config, get_pose2d_fnames)
    cam_videos = defaultdict(list)
    all_cam_names = set()

    for key, (session_path, fnames) in pose_fnames.items():
        for fname in fnames:
            # print(fname)
            vidname = get_video_name(config, fname)
            cname = get_cam_name(config, fname)
            k = (key, session_path, vidname)
            cam_videos[k].append(fname)
            all_cam_names.add(cname)

    all_cam_names = sorted(list(all_cam_names))
    vid_names = sorted(cam_videos.keys())

    all_points = []
    all_scores = []

    for name in tqdm(vid_names, desc='load points', ncols=80):
        (key, session_path, vidname) = name
        fnames = sorted(cam_videos[name])
        cam_names = [get_cam_name(config, f) for f in fnames]
        fname_dict = dict(zip(cam_names, fnames))
        video_folder = os.path.join(session_path, config['pipeline']['videos_raw'])
        offsets_dict = load_offsets_dict(config, cam_names, video_folder)
        out = load_pose2d_fnames(fname_dict, offsets_dict, cam_names)
        points_raw_dict = dict(zip(cam_names, out['points']))
        scores_dict = dict(zip(cam_names, out['scores']))

        _, n_frames, n_joints, _ = out['points'].shape
        points_raw = np.full((len(all_cam_names), n_frames, n_joints, 2), np.nan, 'float')
        scores = np.full((len(all_cam_names), n_frames, n_joints), np.nan, 'float')

        for cnum, cname in enumerate(all_cam_names):
            if cname in points_raw_dict:
                points_raw[cnum] = points_raw_dict[cname]
                scores[cnum] = scores_dict[cname]


        all_points.append(points_raw)
        all_scores.append(scores)

    all_points = np.hstack(all_points)
    all_scores = np.hstack(all_scores)

    return all_points, all_scores, all_cam_names

def process_points_for_calibration(all_points, all_scores):
    """Takes in an array all_points of shape CxFxJx2 and all_scores of shape CxFxJ, where
    C: number of cameras
    F: number of frames
    J: number of joints"""

    assert all_points.shape[3] == 2, \
        "points are not 2 dimensional, or shape of points is wrong: {}".format(all_points.shape)

    n_cams, n_frames, n_joints, _ = all_points.shape

    points = np.copy(all_points).reshape(n_cams, -1, 2)
    scores = all_scores.reshape(n_cams, -1)

    bad = np.isnan(points[:, :, 0])
    scores[bad] = 0

    thres = np.percentile(scores, 90)
    thres = max(min(thres, 0.95), 0.8)
    points[scores < thres] = np.nan

    num_good = np.sum(~np.isnan(points[:, :, 0]), axis=0)
    good = num_good >= 2
    points = points[:, good]

    max_size = int(100e3)

    if points.shape[1] > max_size:
        sample_ixs = np.random.choice(points.shape[1], size=max_size, replace=False)
        points = points[:, sample_ixs]

    return points

def process_session(config, session_path):
    pipeline_calibration_videos = config['pipeline']['calibration_videos']
    pipeline_calibration_results = config['pipeline']['calibration_results']
    video_ext = config['video_extension']

    print(session_path)
    
    calibration_path = find_calibration_folder(config, session_path)

    if calibration_path is None:
        return

    videos = glob(os.path.join(calibration_path,
                               pipeline_calibration_videos,
                               '*.'+video_ext))
    videos = sorted(videos)


    cam_videos = defaultdict(list)
    cam_names = set()
    for vid in videos:
        name = get_cam_name(config, vid)
        cam_videos[name].append(vid)
        cam_names.add(name)

    cam_names = sorted(cam_names)

    video_list = [sorted(cam_videos[cname]) for cname in cam_names]

    outname_base = 'calibration.toml'
    outdir = os.path.join(calibration_path, pipeline_calibration_results)
    outname = os.path.join(outdir, outname_base)

    print(outname)
    skip_calib = False
    init_stuff = True
    error = None

    if os.path.exists(outname):
        cgroup = CameraGroup.load(outname)
        if (not config['calibration']['animal_calibration']) or \
           ('adjusted' in cgroup.metadata and cgroup.metadata['adjusted']):
            return
        else:
            skip_calib = True
            if 'error' in cgroup.metadata:
                error = cgroup.metadata['error']
            else:
                error = None
        init_stuff = False
    elif config['calibration']['calibration_init'] is not None:
        calib_path = os.path.join(config['path'], config['calibration']['calibration_init'])
        print('loading calibration from: {}'.format(calib_path))
        cgroup = CameraGroup.load(calib_path)
        init_stuff = False
        skip_calib = len(videos) == 0
    else:
        if len(videos) == 0:
            print('no videos or calibration file found, continuing...')
            return
        cgroup = CameraGroup.from_names(cam_names, config['calibration']['fisheye'])

    board = get_calibration_board(config)


    if not skip_calib:
        rows_fname = os.path.join(outdir, 'detections.pickle')
        if os.path.exists(rows_fname):
            with open(rows_fname, 'rb') as f:
                all_rows = pickle.load(f)
        else:
            all_rows = cgroup.get_rows_videos(video_list, board)
            with open(rows_fname, 'wb') as f:
                pickle.dump(all_rows, f)

        cgroup.set_camera_sizes_videos(video_list)

        # cgroup.calibrate_rows(all_rows, board,
        #                       init_extrinsics=init_stuff,
        #                       init_intrinsics=init_stuff,
        #                       max_nfev=100, n_iters=2,
        #                       n_samp_iter=100, n_samp_full=300,
        #                       verbose=True)
        error = cgroup.calibrate_rows(all_rows, board,
                                      init_intrinsics=init_stuff, init_extrinsics=init_stuff,
                                      max_nfev=200, n_iters=6,
                                      n_samp_iter=200, n_samp_full=1000,
                                      verbose=True)

    cgroup.metadata['adjusted'] = False
    if error is not None:
        cgroup.metadata['error'] = float(error)
    cgroup.dump(outname)
        
    if config['calibration']['animal_calibration']:
        all_points, all_scores, all_cam_names = load_2d_data(config, calibration_path)
        imgp = process_points_for_calibration(all_points, all_scores)
        # error = cgroup.bundle_adjust(imgp, threshold=10, ftol=1e-4, loss='huber')
        cgroup = cgroup.subset_cameras_names(all_cam_names)
        error = cgroup.bundle_adjust_iter(imgp, ftol=1e-4, n_iters=10,
                                          n_samp_iter=300, n_samp_full=1000,
                                          max_nfev=500,
                                          verbose=True)
        cgroup.metadata['adjusted'] = True
        cgroup.metadata['error'] = float(error)

    cgroup.dump(outname)


calibrate_all = make_process_fun(process_session)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="anipose",
    version="1.1.10",
    author="Lili Karashchuk",
    author_email="krchtchk@gmail.com",
    description="Framework for scalable DeepLabCut based analysis including 3D tracking",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lambdaloop/anipose",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Image Recognition"
    ],
    entry_points={
        'console_scripts': ['anipose=anipose.anipose:cli']
    },
    install_requires=[
        'deeplabcut>=2.0.4.1',
        'aniposelib>=0.4.3',
        'opencv-contrib-python',
        'toml',
        'numpy',
        'scipy',
        'pandas',
        'tqdm',
        'click',
        'scikit-video',
        'flask', 'flask-compress', 'flask-ipban'
    ],
    extras_require={
        'viz':  ["mayavi"]
    }

)



#!/usr/bin/env python3

import numpy as np
from glob import glob
import pandas as pd
import os.path
import cv2
from tqdm import tqdm, trange
from collections import defaultdict
from scipy import signal
import queue
import threading
from datetime import datetime
from ruamel.yaml import YAML
import shutil

from aniposelib.cameras import CameraGroup

from .common import make_process_fun, process_all, get_nframes, \
    get_video_name, get_cam_name, \
    get_video_params, get_video_params_cap, \
    get_data_length, natural_keys, true_basename, find_calibration_folder

from .triangulate import load_pose2d_fnames, load_offsets_dict

def nan_helper(y):
    return np.isnan(y), lambda z: z.nonzero()[0]

def clean_index(df):
    df.index = [x.replace('\\', '/') for x in df.index]
    df.index = [x.replace('labeled-data/', '') for x in df.index]
    return df

def read_frames(caps_2d):
    frames_2d = []
    for cap in caps_2d:
        ret, frame = cap.read()
        if not ret:
            return False, None
        # img = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
        img = frame
        frames_2d.append(img)

    return ret, frames_2d

def get_video_params_cap(cap):
    params = dict()
    params['width'] = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
    params['height'] = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
    params['nframes'] = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    params['fps'] = cap.get(cv2.CAP_PROP_FPS)
    return params


def get_projected_points(bodyparts, pose_fname, cgroup, offsets_dict=None):
    pose_data = pd.read_csv(pose_fname)
    cols = [x for x in pose_data.columns if '_error' in x]


    M = np.identity(3)
    center = np.zeros(3)
    for i in range(3):
        center[i] = np.mean(pose_data['center_{}'.format(i)])
        for j in range(3):
            M[i, j] = np.mean(pose_data['M_{}{}'.format(i, j)])

    bp_dict = dict(zip(bodyparts, range(len(bodyparts))))

    all_points = np.array([np.array(pose_data.loc[:, (bp+'_x', bp+'_y', bp+'_z')])
                           for bp in bodyparts])

    all_errors = np.array([np.array(pose_data.loc[:, bp+'_error'])
                           for bp in bodyparts])

    n_joints, n_frames, _ = all_points.shape
    n_cams = len(cgroup.cameras)

    all_points_flat = all_points.reshape(-1, 3)
    all_points_flat_t = (all_points_flat + center).dot(np.linalg.inv(M.T))

    points_2d_proj_flat = cgroup.project(all_points_flat_t)
    points_2d_proj = points_2d_proj_flat.reshape(n_cams, n_joints, n_frames, 2)

    cam_names = cgroup.get_names()
    if offsets_dict is None:
        offsets_dict = dict([(cname, (0,0)) for cname in cam_names])

    for cix, cname in enumerate(cam_names):
        offset = offsets_dict[cname]
        dx, dy = offset[0], offset[1]
        points_2d_proj[cix, :, :, 0] -= dx
        points_2d_proj[cix, :, :, 1] -= dy

    return points_2d_proj.swapaxes(1, 2)

def get_pose2d_fnames(config, session_path):
    pipeline_pose = config['pipeline']['pose_2d']
    fnames = glob(os.path.join(session_path, pipeline_pose, '*.h5'))
    return session_path, fnames

def get_videos_fnames(config, session_path):
    pipeline_raw = config['pipeline']['videos_raw']
    ext = config['video_extension']
    fnames = glob(os.path.join(session_path, pipeline_raw, '*.' + ext))
    return session_path, fnames


def load_2d_data(config):
    pose_fnames = process_all(config, get_pose2d_fnames)
    cam_videos = defaultdict(list)

    for key, (session_path, fnames) in pose_fnames.items():
        for fname in fnames:
            # print(fname)
            vidname = get_video_name(config, fname)
            k = (key, session_path, vidname)
            cam_videos[k].append(fname)

    vid_names = sorted(cam_videos.keys())

    all_points = []
    all_scores = []
    all_proj = []
    all_fnames = []
    calib_fnames = []

    for name in tqdm(vid_names, desc='load points', ncols=80):
    # for name in vid_names:
        (key, session_path, vidname) = name
        fnames = sorted(cam_videos[name])
        cam_names = [get_cam_name(config, f) for f in fnames]
        fname_dict = dict(zip(cam_names, fnames))

        cgroup = None
        calib_folder = find_calibration_folder(config, session_path)
        if calib_folder is not None:
            calib_fname = os.path.join(calib_folder,
                                       config['pipeline']['calibration_results'],
                                       'calibration.toml')
            if os.path.exists(calib_fname):
                cgroup = CameraGroup.load(calib_fname)

        pose_fname = os.path.join(session_path, config['pipeline']['pose_3d'],
                                  vidname+'.csv')

        if cgroup is None or not os.path.exists(pose_fname):
            continue

        calib_fnames.append(calib_fname)

        video_folder = os.path.join(session_path, config['pipeline']['videos_raw'])
        offsets_dict = load_offsets_dict(config, cam_names, video_folder)
        out = load_pose2d_fnames(fname_dict)
        points_raw = out['points']
        scores = out['scores']
        bodyparts = out['bodyparts']

        vid_fnames = [os.path.join(session_path,
                                   config['pipeline']['videos_raw'],
                                   true_basename(f) + '.' + config['video_extension'])
                      for f in fnames]

        points_proj = get_projected_points(
            bodyparts, pose_fname, cgroup, offsets_dict)

        all_points.append(points_raw)
        all_scores.append(scores)
        all_proj.append(points_proj)
        all_fnames.append(vid_fnames)


    out = {
        'points': all_points,
        'scores': all_scores,
        'proj': all_proj,
        'fnames': all_fnames,
        'cam_names': cam_names,
        'calib_fnames': calib_fnames,
        'bodyparts': bodyparts
    }
    return out

def get_all_videos_fnames(config):
    vids_fnames = process_all(config, get_videos_fnames)
    cam_videos = defaultdict(list)

    for key, (session_path, fnames) in vids_fnames.items():
        for fname in fnames:
            # print(fname)
            vidname = get_video_name(config, fname)
            k = (key, session_path, vidname)
            cam_videos[k].append(fname)

    vid_names = sorted(cam_videos.keys(), key=lambda x: natural_keys(x[2]))

    all_fnames = []
    calib_fnames = []

    for name in tqdm(vid_names, desc='get videos', ncols=80):
        (key, session_path, vidname) = name
        fnames = sorted(cam_videos[name], key=natural_keys)
        cam_names = [get_cam_name(config, f) for f in fnames]

        cgroup = None
        calib_folder = find_calibration_folder(config, session_path)
        if calib_folder is not None:
            calib_fname = os.path.join(calib_folder,
                                       config['pipeline']['calibration_results'],
                                       'calibration.toml')
            if os.path.exists(calib_fname):
                cgroup = CameraGroup.load(calib_fname)

        # pose_fname = os.path.join(session_path, config['pipeline']['pose_3d'],
        #                           vidname+'.csv')

        if cgroup is None: # or not os.path.exists(pose_fname):
            continue

        calib_fnames.append(calib_fname)
        all_fnames.append(fnames)
        
    normal_count = max([len(x) for x in all_fnames])
    bad_num = np.sum([len(x) != normal_count for x in all_fnames])
    if bad_num > 0:
        print('W: ignored {} sets of videos with inconsistent number of cameras'.format(bad_num))
        all_fnames = [x for x in all_fnames if len(x) == normal_count]

    out = {
        'fnames': all_fnames,
        'calib_fnames': calib_fnames,
        'cam_names': cam_names
    }
    return out

def extract_frames_random(config, num_frames_pick=250, name=None):
    d = get_all_videos_fnames(config)
    all_fnames = d['fnames']
    cam_names = d['cam_names']
    calib_fnames = d['calib_fnames']

    main_folder = os.path.basename(os.getcwd())

    n_cams = len(cam_names)

    # model_folder = config['model_folder']
    # dlc_config_fname = os.path.join(model_folder, 'config.yaml')

    # yaml = YAML(typ='rt')
    # with open(dlc_config_fname, 'r') as f:
    #     dlc_config = yaml.load(f)

    vidnums = []
    framenums = []

    for vnum, fnames in enumerate(all_fnames):
        num_frames = np.inf
        for fname in fnames:
            params = get_video_params(fname)
            num_frames = min(num_frames, params['nframes'])
            
        vidnums.append(np.ones(num_frames, dtype='int64')*vnum)
        framenums.append(np.arange(num_frames))

    vidnums = np.hstack(vidnums)
    framenums = np.hstack(framenums)

    num_total = len(vidnums)
    
    indexes = np.arange(num_total)
    np.random.shuffle(indexes)
    
    check = np.ones(num_total, dtype='bool')

    picked = []
    for ix in indexes:
        vidnum = vidnums[ix]
        framenum = framenums[ix]
        if not check[ix]:
            continue

        similar = (vidnum == vidnums) & (np.abs(framenums - framenum) <= 10)
        check[similar] = False

        picked.append( (vidnum, framenum) )
        if len(picked) >= num_frames_pick:
            break

    picked = sorted(picked)

    nd = int(np.log10(num_frames_pick) + 1)
    img_format = 'img{:0' + str(nd) +'d}.png'
    images = [img_format.format(i) for i in range(num_frames_pick)]

    if name is None:
        folder_base = '{}_{}_{}_random'.format(
            config['project'], main_folder, datetime.now().strftime('%Y-%m-%d_%H-%M'))
    else:
        folder_base = name

    folders = []
    metas = []
    indexes = []
    for cnum in range(n_cams):
        folder = folder_base + '--' + cam_names[cnum]
        images_cur = [os.path.join('labeled-data', folder_base, folder, img) for img in images]
        indexes.append(images_cur)

        meta = pd.DataFrame(columns=['img', 'calib', 'video', 'framenum'])
        metas.append(meta)

        # full_folder = os.path.join(model_folder, 'labeled-data', folder)
        full_folder = os.path.join('labeled-data', folder_base, folder)
        os.makedirs(full_folder, exist_ok=True)
        folders.append(full_folder)

    caps = []
    vidnum_cur = None

    max_width = defaultdict(int)
    max_height = defaultdict(int)

    rowcount = 0

    for imgnum, (vidnum, framenum) in enumerate(tqdm(picked, ncols=70)):
        # print(vidnum, framenum)
 
        if vidnum_cur != vidnum:
            for cap in caps:
                cap.release()
            caps = [cv2.VideoCapture(f) for f in all_fnames[vidnum]]
            vidnum_cur = vidnum
            framenum_cur = 0

        while framenum_cur < framenum:
            ret, frames = read_frames(caps)
            framenum_cur += 1
            if not ret:
                break

        ret, frames = read_frames(caps)
        framenum_cur += 1
        if not ret:
            continue

        for cnum in range(n_cams):
            row = indexes[cnum][imgnum]

            metas[cnum].loc[rowcount, 'img'] = row
            metas[cnum].loc[rowcount, 'calib'] = calib_fnames[vidnum]
            metas[cnum].loc[rowcount, 'video'] = all_fnames[vidnum][cnum]
            metas[cnum].loc[rowcount, 'framenum'] = framenum

            cv2.imwrite(row, frames[cnum])
            height, width, _  = frames[cnum].shape
            max_height[cnum] = max(max_height[cnum], height)
            max_width[cnum] = max(max_width[cnum], width)

        rowcount += 1

    for cnum, folder in enumerate(folders):
        metas[cnum].to_csv(os.path.join(folder, 'anipose_metadata.csv'), index=False)

        # key = true_basename(folder) + '.avi'
        # dlc_config['video_sets'][key] = {
        #     'crop': '0, {}, 0, {}'.format(max_height[cnum], max_width[cnum])
        # }

    config_path = os.path.join(config['path'], 'config.toml')
    if os.path.exists(config_path):
        shutil.copy(config_path, os.path.join('labeled-data', folder_base, 'config.toml'))

    calib_path = os.path.join(config['path'],
                              config['pipeline']['calibration_results'],
                              'calibration.toml')
    if os.path.exists(calib_path):
        shutil.copy(calib_path, os.path.join('labeled-data', folder_base, 'calibration.toml'))

    # with open(dlc_config_fname, 'w') as f:
    #     yaml.dump(dlc_config, f)

    
POSSIBLE_MODES = ['good', 'bad', 'random']
def extract_frames_picked(config, mode='bad', num_frames_pick=250, scorer=None, name=None):
    if mode not in POSSIBLE_MODES:
        raise ValueError(
            'extract_frames_picked needs mode to be one of {}, but received "{}"'
            .format(POSSIBLE_MODES, mode))

    print('loading tracked data...')
    d = load_2d_data(config)

    print('getting {} frames...'.format(mode))
    points = d['points']
    proj = d['proj']
    scores = d['scores']
    fnames = d['fnames']
    cam_names = d['cam_names']
    calib_fnames = d['calib_fnames']
    bodyparts = d['bodyparts']

    main_folder = os.path.basename(os.getcwd())

    model_folder = config['model_folder']
    dlc_config_fname = os.path.join(model_folder, 'config.yaml')


    if scorer is None:
        # try to read it from yaml file
        if os.path.exists(dlc_config_fname):
            yaml = YAML(typ='rt')
            with open(dlc_config_fname, 'r') as f:
                dlc_config = yaml.load(f)
            scorer = dlc_config['scorer']
        else:
            scorer = 'default'

    nums = [p.shape[1] for p in points]
    num_total = np.sum(nums)
    n_cams = proj[0].shape[0]

    vidnums = np.zeros(num_total, dtype='int64')
    framenums = np.zeros(num_total, dtype='int64')
    errors = np.zeros(num_total, dtype='float64')

    start = 0
    for vnum, num_frames in enumerate(nums):
        a = start
        b = start + num_frames

        errors_cur = np.linalg.norm(proj[vnum] - points[vnum], axis=3)
        errors_mean = np.mean(np.mean(errors_cur, axis=2), axis=0)

        vidnums[a:b] = vnum
        framenums[a:b] = np.arange(num_frames)
        errors[a:b] = errors_mean
        start += num_frames

    good = np.isfinite(errors)
    errors[~good] = np.max(errors[good])*0.5

    if mode == 'bad':
        log_errors = np.log(errors+0.1)
        log_errors = np.clip(log_errors, -np.inf, np.percentile(log_errors, 85))
        error_percent = np.max(log_errors) - np.percentile(log_errors, 60)
        noise = np.random.uniform(0, error_percent, size=errors.shape)
        indexes = np.argsort(-log_errors + noise)
    elif mode == 'good':
        error_percent = np.percentile(errors, 20)
        noise = np.random.uniform(-error_percent, error_percent, size=errors.shape)
        indexes = np.argsort(errors + noise)
    elif mode == 'random':
        indexes = np.arange(len(errors))
        np.random.shuffle(indexes)

    check = np.ones(num_total, dtype='bool')

    picked = []
    for ix in indexes:
        vidnum = vidnums[ix]
        framenum = framenums[ix]
        if not check[ix]:
            continue

        similar = (vidnum == vidnums) & (np.abs(framenums - framenum) <= 10)
        check[similar] = False

        picked.append( (vidnum, framenum) )
        if len(picked) >= num_frames_pick:
            break

    picked = sorted(picked)


    columns = pd.MultiIndex.from_product(
        [[scorer], bodyparts, ['x', 'y']],
        names=['scorer', 'bodyparts', 'coords'])

    nd = int(np.log10(num_frames_pick) + 1)
    img_format = 'img{:0' + str(nd) +'d}.png'
    images = [img_format.format(i) for i in range(num_frames_pick)]

    # folder_base = datetime.now().strftime('%Y-%m-%d--%H-%M')
    if name is None:
        folder_base = '{}_{}_{}_{}'.format(
            config['project'], main_folder, datetime.now().strftime('%Y-%m-%d_%H-%M'), mode)
    else:
        folder_base = name

    folders = []
    douts = []
    metas = []
    indexes = []
    for cnum in range(n_cams):
        folder = folder_base + '--' + cam_names[cnum]
        images_cur = [os.path.join('labeled-data', folder_base, folder, img) for img in images]
        indexes.append(images_cur)

        dout = pd.DataFrame(columns=columns)
        dout = dout.astype('float64')
        douts.append(dout)

        meta = pd.DataFrame(columns=['img', 'calib', 'video', 'framenum'])
        metas.append(meta)

        # full_folder = os.path.join(model_folder, 'labeled-data', folder)
        full_folder = os.path.join('labeled-data', folder_base, folder)
        os.makedirs(full_folder, exist_ok=True)
        folders.append(full_folder)

    caps = []
    vidnum_cur = None

    max_width = defaultdict(int)
    max_height = defaultdict(int)

    rowcount = 0

    for imgnum, (vidnum, framenum) in enumerate(tqdm(picked, ncols=70)):
        # print(vidnum, framenum)

        if vidnum_cur != vidnum:
            for cap in caps:
                cap.release()
            caps = [cv2.VideoCapture(f) for f in fnames[vidnum]]
            vidnum_cur = vidnum
            framenum_cur = 0

        while framenum_cur < framenum:
            ret, frames = read_frames(caps)
            framenum_cur += 1
            if not ret:
                break

        ret, frames = read_frames(caps)
        framenum_cur += 1
        if not ret:
            continue

        pred = proj[vidnum][:, framenum]
        sc = scores[vidnum][:, framenum]
        pred[sc < 0.3] = np.nan

        for cnum in range(n_cams):
            row = indexes[cnum][imgnum]
            row_x = row.replace(folder_base + '/', '')
            douts[cnum].loc[row_x, (scorer, bodyparts, 'x')] = pred[cnum, :, 0]
            douts[cnum].loc[row_x, (scorer, bodyparts, 'y')] = pred[cnum, :, 1]

            metas[cnum].loc[rowcount, 'img'] = row
            metas[cnum].loc[rowcount, 'calib'] = calib_fnames[vidnum]
            metas[cnum].loc[rowcount, 'video'] = fnames[vidnum][cnum]
            metas[cnum].loc[rowcount, 'framenum'] = framenum

            cv2.imwrite(row, frames[cnum])
            height, width, _  = frames[cnum].shape
            max_height[cnum] = max(max_height[cnum], height)
            max_width[cnum] = max(max_width[cnum], width)

        rowcount += 1

    for cnum, folder in enumerate(folders):
        douts[cnum].to_hdf(os.path.join(folder, 'CollectedData_' + scorer + '.h5'),
                      key='df_with_missing', format='table', mode='w')
        douts[cnum].to_csv(os.path.join(folder, 'CollectedData_' + scorer + '.csv'))

        metas[cnum].to_csv(os.path.join(folder, 'anipose_metadata.csv'), index=False)

        key = true_basename(folder) + '.avi'
        # dlc_config['video_sets'][key] = {
        #     'crop': '0, {}, 0, {}'.format(max_height[cnum], max_width[cnum])
        # }

    data_concat = pd.concat([clean_index(d.copy()) for d in douts])
    data_concat.to_csv(os.path.join('labeled-data', folder_base, 'CollectedData.csv'), index=True)

    config_path = os.path.join(config['path'], 'config.toml')
    if os.path.exists(config_path):
        shutil.copy(config_path, os.path.join('labeled-data', folder_base, 'config.toml'))

    calib_path = os.path.join(config['path'],
                              config['pipeline']['calibration_results'],
                              'calibration.toml')
    if os.path.exists(calib_path):
        shutil.copy(calib_path, os.path.join('labeled-data', folder_base, 'calibration.toml'))

    # calibration_path =
    # with open(dlc_config_fname, 'w') as f:
    #     yaml.dump(dlc_config, f)



#!/usr/bin/env python3

from mayavi import mlab
mlab.options.offscreen = True

import numpy as np
from glob import glob
import pandas as pd
import os.path
import cv2
import sys
import skvideo.io
from tqdm import tqdm, trange
import sys
from collections import defaultdict
from matplotlib.pyplot import get_cmap

from .common import make_process_fun, get_nframes, get_video_name, get_video_params, get_data_length, natural_keys


def connect(points, bps, bp_dict, color):
    ixs = [bp_dict[bp] for bp in bps]
    return mlab.plot3d(points[ixs, 0], points[ixs, 1], points[ixs, 2],
                       np.ones(len(ixs)), reset_zoom=False,
                       color=color, tube_radius=None, line_width=10)

def connect_all(points, scheme, bp_dict, cmap):
    lines = []
    for i, bps in enumerate(scheme):
        line = connect(points, bps, bp_dict, color=cmap(i)[:3])
        lines.append(line)
    return lines

def update_line(line, points, bps, bp_dict):
    ixs = [bp_dict[bp] for bp in bps]
    # ixs = [bodyparts.index(bp) for bp in bps]
    new = np.vstack([points[ixs, 0], points[ixs, 1], points[ixs, 2]]).T
    line.mlab_source.points = new

def update_all_lines(lines, points, scheme, bp_dict):
    for line, bps in zip(lines, scheme):
        update_line(line, points, bps, bp_dict)



def visualize_labels(config, labels_fname, outname, fps=300):

    try:
        scheme = config['labeling']['scheme']
    except KeyError:
        scheme = []

    data = pd.read_csv(labels_fname)
    cols = [x for x in data.columns if '_error' in x]

    if len(scheme) == 0:
        bodyparts = [c.replace('_error', '') for c in cols]
    else:
        bodyparts = sorted(set([x for dx in scheme for x in dx]))

    bp_dict = dict(zip(bodyparts, range(len(bodyparts))))

    all_points = np.array([np.array(data.loc[:, (bp+'_x', bp+'_y', bp+'_z')])
                           for bp in bodyparts], dtype='float64')

    all_errors = np.array([np.array(data.loc[:, bp+'_error'])
                           for bp in bodyparts], dtype='float64')

    all_scores = np.array([np.array(data.loc[:, bp+'_score'])
                           for bp in bodyparts], dtype='float64')

    all_ncams = np.array([np.array(data.loc[:, bp+'_ncams'])
                          for bp in bodyparts], dtype='float64')


    if config['triangulation']['optim']:
        all_errors[np.isnan(all_errors)] = 0
    else:
        all_errors[np.isnan(all_errors)] = 10000
    good = (all_errors < 100)
    all_points[~good] = np.nan

    not_enough_points = np.mean(all_ncams >= 2, axis=1) < 0.2
    all_points[not_enough_points] = np.nan

    all_points_flat = all_points.reshape(-1, 3)
    check = ~np.isnan(all_points_flat[:, 0])

    if np.sum(check) < 10:
        print('too few points to plot, skipping...')
        return

    low, high = np.percentile(all_points_flat[check], [5, 95], axis=0)

    nparts = len(bodyparts)
    framedict = dict(zip(data['fnum'], data.index))

    writer = skvideo.io.FFmpegWriter(outname, inputdict={
        # '-hwaccel': 'auto',
        '-framerate': str(fps),
    }, outputdict={
        '-vcodec': 'h264', '-qp': '28', '-pix_fmt': 'yuv420p'
    })

    cmap = get_cmap('tab10')


    points = np.copy(all_points[:, 20])
    points[0] = low
    points[1] = high

    s = np.arange(points.shape[0])
    good = ~np.isnan(points[:, 0])

    fig = mlab.figure(bgcolor=(1,1,1), size=(500,500))
    fig.scene.anti_aliasing_frames = 2

    low, high = np.percentile(points[good, 0], [10,90])
    scale_factor = (high - low) / 12.0

    mlab.clf()
    pts = mlab.points3d(points[:, 0], points[:, 1], points[:, 2], s,
                        color=(0.8, 0.8, 0.8),
                        scale_mode='none', scale_factor=scale_factor)
    lines = connect_all(points, scheme, bp_dict, cmap)
    mlab.orientation_axes()

    view = list(mlab.view())

    mlab.view(focalpoint='auto', distance='auto')

    for framenum in trange(data.shape[0], ncols=70):
        fig.scene.disable_render = True

        if framenum in framedict:
            points = all_points[:, framenum]
        else:
            points = np.ones((nparts, 3))*np.nan

        s = np.arange(points.shape[0])
        good = ~np.isnan(points[:, 0])

        new = np.vstack([points[:, 0], points[:, 1], points[:, 2]]).T
        pts.mlab_source.points = new
        update_all_lines(lines, points, scheme, bp_dict)

        fig.scene.disable_render = False

        img = mlab.screenshot()

        mlab.view(*view, reset_roll=False)

        writer.writeFrame(img)

    mlab.close(all=True)
    writer.close()



def process_session(config, session_path, filtered=False):
    pipeline_videos_raw = config['pipeline']['videos_raw']

    if filtered:
        pipeline_videos_labeled_3d = config['pipeline']['videos_labeled_3d_filter']
        pipeline_3d = config['pipeline']['pose_3d_filter']
    else:
        pipeline_videos_labeled_3d = config['pipeline']['videos_labeled_3d']
        pipeline_3d = config['pipeline']['pose_3d']

    video_ext = config['video_extension']

    vid_fnames = glob(os.path.join(session_path,
                                   pipeline_videos_raw, "*."+video_ext))
    orig_fnames = defaultdict(list)
    for vid in vid_fnames:
        vidname = get_video_name(config, vid)
        orig_fnames[vidname].append(vid)

    labels_fnames = glob(os.path.join(session_path,
                                      pipeline_3d, '*.csv'))
    labels_fnames = sorted(labels_fnames, key=natural_keys)


    outdir = os.path.join(session_path, pipeline_videos_labeled_3d)

    if len(labels_fnames) > 0:
        os.makedirs(outdir, exist_ok=True)

    for fname in labels_fnames:
        basename = os.path.basename(fname)
        basename = os.path.splitext(basename)[0]

        out_fname = os.path.join(outdir, basename+'.mp4')

        if os.path.exists(out_fname) and \
           abs(get_nframes(out_fname) - get_data_length(fname)) < 100:
            continue
        print(out_fname)

        some_vid = orig_fnames[basename][0]
        params = get_video_params(some_vid)

        visualize_labels(config, fname, out_fname, params['fps'])


label_videos_3d_all = make_process_fun(process_session, filtered=False)
label_videos_3d_filtered_all = make_process_fun(process_session, filtered=True)
